<?php
/**
 * YetiForce system update package file.
 *
 * @package   YetiForce.UpdatePackages
 *
 * @copyright YetiForce S.A.
 * @license   YetiForce Public License 7.0 (licenses/LicenseEN.txt or yetiforce.com)
 * @author    Radosław Skrzypczak <r.skrzypczak@yetiforce.com>
 */

/**
 * YetiForce system update package class.
 */
class YetiForceUpdate
{
    /** @var \vtlib\PackageImport */
    public $package;

    /** @var string[] Fields to delete. */
    public $filesToDelete = [];

    /** @var string */
    private $logFile = 'cache/logs/updateLogsTrace.log';

    /** @var object Module Meta XML File (Parsed). */
    private $moduleNode;

    /** @var DbImporter */
    private $importer;

    /** @var string[] Errors. */
    private $error = [];

    /**
     * Constructor.
     *
     * @param object $moduleNode
     */
    public function __construct($moduleNode)
    {
        $this->moduleNode = $moduleNode;
        $this->filesToDelete = require_once 'deleteFiles.php';
    }

    /**
     * Log.
     *
     * @param string $message Logs.
     */
    private function log(string $message): void
    {
        $fp = fopen($this->logFile, 'a+');
        fwrite($fp, $message . PHP_EOL);
        fclose($fp);
        if (false !== stripos($message, '[ERROR]')) {
            $this->error[] = $message;
        }
    }

    /**
     * Pre update.
     */
    public function preupdate(): bool
    {
        $minTime = 600;
        $maxExecutionTime = ini_get('max_execution_time');
        $maxInputTime = ini_get('max_input_time');

        if (version_compare(PHP_VERSION, '8.1', '<')) {
            $this->package->_errorText = 'The server configuration is not compatible with the requirements of the upgrade package.' . PHP_EOL;
            $this->package->_errorText .= 'Please have a look at the list of errors:' . PHP_EOL . PHP_EOL;
            $this->package->_errorText .= 'Wrong PHP version, recommended version >= 8.1';
            return false;
        }

        if ((is_numeric($maxExecutionTime) && 0 != $maxExecutionTime && $maxExecutionTime < $minTime) || (is_numeric($maxInputTime) && $maxInputTime > 0 && $maxInputTime < $minTime)) {
            $this->package->_errorText = 'The server configuration is not compatible with the requirements of the upgrade package.' . PHP_EOL;
            $this->package->_errorText .= 'Please have a look at the list of errors:';
            if (0 != $maxExecutionTime && $maxExecutionTime < $minTime) {
                $this->package->_errorText .= PHP_EOL . 'max_execution_time = ' . $maxExecutionTime . ' < ' . $minTime;
            }
            if ($maxInputTime > 0 && $maxInputTime < $minTime) {
                $this->package->_errorText .= PHP_EOL . 'max_input_time = ' . $maxInputTime . ' < ' . $minTime;
            }
            return false;
        }

        if(!\function_exists('openssl_encrypt') || !extension_loaded('openssl')) {
            $this->package->_errorText = 'The server configuration is not compatible with the requirements: openssl extension is required.' . PHP_EOL;
            return false;
        }
        return true;
    }

    /**
     * Update.
     */
    public function update(): void
    {
        $start = microtime(true);
        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s'));

        $this->importer = new \App\Db\Importer();
        try {
            $this->importer->loadFiles(__DIR__ . '/dbscheme');
            $this->importer->checkIntegrity(false);
            $this->importer->dropTable([
                'vtiger_userscf',
            ]);

            $this->importer->dropForeignKeys([
               'roundcube_users_autologin_ibfk_1' =>  'roundcube_users_autologin',
               'u_yf_mail_quantities_ibfk_1' =>  'u_yf_mail_quantities',
            ]);

            $this->preSchemaUpdateData();
            $this->importer->updateScheme();
            $this->importer->refreshSchema();
            $this->importer->postUpdate();
            $this->addModules(['ReportTemplate']);
            $this->importer->logs(false);
        } catch (\Throwable $ex) {
            $this->log($ex->getMessage() . '|' . $ex->getTraceAsString());
            $this->importer->logs(false);
            throw $ex;
        }

        $this->importer->refreshSchema();
        $this->importer->checkIntegrity(true);

        $this->addFields($this->getFields(0));
        $this->setRelations();

        $this->migrateLdap();
        $this->changeUitypeForPicklist();
        $this->updateData();

        $this->smtp();
        $this->migrateMailServer();
        $this->migrateMailAccounts();
        $this->migrateMailConfig();

        $this->addActionMapping();
        $this->modulePackage();
        $this->migrateMap();
        $this->inventoryDefaults();
        $this->updateSettingsMenu();

        $this->importer->dropTable([
                'vtiger_ossmailscanner_config',
                'vtiger_ossmailscanner_folders_uid',
            ]);

        $this->createConfigFiles();
        (new \Settings_Menu_Record_Model())->refreshMenuFiles();
        \App\Module::createModuleMetaFile();
        \App\UserPrivilegesFile::recalculateAll();
        $this->importer->dropColumns([
            ['vtiger_tab', 'parent'],
        ]);
        $this->importer->logs(false);
        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s') . ' | ' . round((microtime(true) - $start) / 60, 2) . ' min');
    }

    public function addBlock()
    {
        $start = microtime(true);
        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s'));

        $moduleModel = \Vtiger_Module_Model::getInstance('Users');
        if (!($moduleModel->getBlocks()['LBL_PBX'] ?? null)) {
            $blockInstance = new \vtlib\Block();
            $data = ['label' => 'LBL_PBX', 'showtitle' => 0, 'visible' => 0, 'increateview' => 0, 'ineditview' => 0, 'indetailview' => 0, 'display_status' => 1, 'iscustom' => 1, 'icon' => 'fa-solid fa-square-phone', 'sequence' => 9];
            foreach ($data as $key => $value) {
                $blockInstance->{$key} = $value;
            }
            $moduleModel->addBlock($blockInstance);
            $this->log('  [INFO] Added new block: LBL_PBX');

            $messageField = $moduleModel->getFieldByName('user_pbx');
            Settings_LayoutEditor_Block_Model::updateFieldSequenceNumber([[
                'block' => $blockInstance->id,
                'fieldid' => $messageField->getId(),
                'sequence' => 1,
            ]]);

            $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s') . ' | ' . round((microtime(true) - $start) / 60, 2) . ' min');
        }
    }
    /**
     * Add modules.
     *
     * @param string[] $modules
     */
    private function addModules(array $modules)
    {
        $start = microtime(true);
        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s'));
        $command = \App\Db::getInstance()->createCommand();
        foreach ($modules as $moduleName) {
            if (file_exists(__DIR__ . '/' . $moduleName . '.xml') && !\vtlib\Module::getInstance($moduleName)) {
                $importInstance = new \vtlib\PackageImport();
                $importInstance->_modulexml = simplexml_load_file('cache/updates/updates/' . $moduleName . '.xml');
                $importInstance->importModule();
                $command->update('vtiger_tab', ['customized' => 0], ['name' => $moduleName])->execute();
                if ($tabId = (new \App\Db\Query())->select(['tabid'])->from('vtiger_tab')->where(['name' => $moduleName])->scalar()) {
                    \CRMEntity::getInstance('ModTracker')->enableTrackingForModule($tabId);
                }
                $this->log('  [INFO] Add module:' . $moduleName);
            } else {
                $this->log('  [INFO] Module exist: ' . $moduleName);
            }
        }
        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s') . ' | ' . round((microtime(true) - $start) / 60, 2) . ' min');
    }

    private function setRelations()
    {
        $start = microtime(true);
        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s'));
        $dbCommand = \App\Db::getInstance()->createCommand();

        $ralations = [
            ['type' => 'add', 'data' => [684, 'Accounts', 'BankAccounts' ,'getDependentsList',45,'BankAccounts',0,'ADD',0,0,0,'RelatedTab','related_to',null]],
        ];

        foreach ($ralations as $relation) {
            [, $moduleName, $relModuleName, $name, $sequence, $label, $presence, $actions, $favorites, $creatorDetail, $relationComment, $viewType, $fieldName,$customView] = $relation['data'];
            $tabid = \App\Module::getModuleId($moduleName);
            $relTabid = \App\Module::getModuleId($relModuleName);
            $where = ['tabid' => $tabid, 'related_tabid' => $relTabid, 'name' => $name];
            $isExists = (new \App\Db\Query())->from('vtiger_relatedlists')->where($where)->exists();
            if (!$isExists && 'add' === $relation['type']) {
                $dbCommand->insert('vtiger_relatedlists', [
                    'tabid' => $tabid,
                    'related_tabid' => $relTabid,
                    'name' => $name,
                    'sequence' => $sequence,
                    'label' => $label,
                    'presence' => $presence,
                    'actions' => $actions,
                    'favorites' => $favorites,
                    'creator_detail' => $creatorDetail,
                    'relation_comment' => $relationComment,
                    'view_type' => $viewType,
                    'field_name' => $fieldName,
                ])->execute();
            } elseif ('update' === $relation['type'] && ($isExists || (!$isExists && isset($relation['where']['name']) && (new \App\Db\Query())->from('vtiger_relatedlists')->where(['tabid' => $tabid, 'related_tabid' => $relTabid])->exists()))) {
                $where = $relation['where'] ?? $where;
                $dbCommand->update('vtiger_relatedlists', [
                    'name' => $name,
                    'sequence' => $sequence,
                    'label' => $label,
                    'presence' => $presence,
                    'actions' => $actions,
                    'favorites' => $favorites,
                    'creator_detail' => $creatorDetail,
                    'relation_comment' => $relationComment,
                    'view_type' => $viewType,
                    'field_name' => $fieldName,
                ], $where)->execute();
            }
        }
        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s') . ' | ' . round((microtime(true) - $start) / 60, 2) . ' mim.');
    }

    private function getFields(int $index)
    {
        $fields = [];
        $importerType = new \App\Db\Importers\Base();
        switch ($index) {
            case 0:
                $fields = [
                    [29, 501, 'user_pbx', 'vtiger_users', 1, 86, 'user_pbx', 'FL_USER_PBX', 0, 2, '', '-32768,32767', 0, 80, 1, 'I~O', 1, 0, 'BAS', 1, '', 0, '', null, 0, 0, 0, 0, '', '', 'type' => $importerType->smallInteger(5)->notNull()->defaultValue(-1),'blockLabel' => 'LBL_PBX', 'blockData' => ['label'=>'LBL_PBX','showtitle'=>0,'visible'=>0,'increateview'=>0,'ineditview'=>0,'indetailview'=>0,'display_status'=>1,'iscustom'=>1,'icon'=>'fa-solid fa-square-phone'], 'moduleName' => 'Users'],
                    [130, 2914, 'coordinates', 'u_yf_locations', 1, 331, 'coordinates', 'FL_COORDINATES', 0, 2, '', '100', 10, 456, 1, 'V~O', 1, 0, 'BAS', 1, '', 0, '{"showType":"1","type":"decimal","showMap":"1","showLocation":"1"}', null, 0, 0, 0, 0, '', '', 'type' => $importerType->stringType(100),'blockLabel' => 'LBL_LOCATIONS_INFORMATION', 'blockData' => ['label'=>'LBL_LOCATIONS_INFORMATION','showtitle'=>0,'visible'=>0,'increateview'=>0,'ineditview'=>0,'indetailview'=>0,'display_status'=>1,'iscustom'=>0,'icon'=>null], 'moduleName' => 'Locations'],
                    [29, 3139, 'mail_popup', 'vtiger_users', 1, 56, 'mail_popup', 'FL_MAIL_POPUP', 0, 2, '', '-128,127', 4, 452, 1, 'C~O', 1, 0, 'BAS', 1, '', 0, '', null, 0, 0, 0, 0, '', '', 'type' => $importerType->tinyInteger(1)->defaultValue(1),'blockLabel' => 'LBL_USER_MAIL', 'blockData' => ['label'=>'LBL_USER_MAIL','showtitle'=>0,'visible'=>0,'increateview'=>0,'ineditview'=>0,'indetailview'=>0,'display_status'=>2,'iscustom'=>0,'icon'=>'fas fa-mail-bulk'], 'moduleName' => 'Users'],
                    [45, 3143, 'sms_provider_id', 'vtiger_smsnotifier', 1, 90, 'sms_provider_id', 'FL_SMS_PROVIDER', 0, 2, '', null, 5, 110, 1, 'I~O', 1, 0, 'BAS', 1, '', 0, '', null, 0, 0, 0, 0, '', '', 'type' => $importerType->integer(10),'blockLabel' => 'LBL_SMSNOTIFIER_INFORMATION', 'blockData' => ['label'=>'LBL_SMSNOTIFIER_INFORMATION','showtitle'=>0,'visible'=>0,'increateview'=>0,'ineditview'=>0,'indetailview'=>0,'display_status'=>2,'iscustom'=>0,'icon'=>null], 'moduleName' => 'SMSNotifier'],
                    [6, 3145, 'account_short_name', 'vtiger_account', 1, 1, 'account_short_name', 'FL_ACCOUNT_SHORT_NAME', 0, 0, '', '255', 12, 9, 1, 'V~M', 1, 0, 'BAS', 1, '', 0, '', null, 0, 0, 0, 0, '', '', 'type' => $importerType->stringType()->defaultValue(''),'blockLabel' => 'LBL_ACCOUNT_INFORMATION', 'blockData' => ['label'=>'LBL_ACCOUNT_INFORMATION','showtitle'=>0,'visible'=>0,'increateview'=>0,'ineditview'=>0,'indetailview'=>0,'display_status'=>2,'iscustom'=>0,'icon'=>null], 'moduleName' => 'Accounts'],
                    [133, 3146, 'related_to', 'u_yf_bankaccounts', 1, 10, 'related_to', 'FL_RELATED_TO', 0, 0, '', '4294967295', 8, 470, 1, 'I~O', 2, 0, 'BAS', 1, '', 0, '', null, 0, 0, 0, 0, '', '', 'type' => $importerType->integer(10)->unsigned()->notNull()->defaultValue(0),'blockLabel' => 'LBL_BASIC_INFORMATION', 'blockData' => ['label'=>'LBL_BASIC_INFORMATION','showtitle'=>0,'visible'=>0,'increateview'=>0,'ineditview'=>0,'indetailview'=>0,'display_status'=>1,'iscustom'=>0,'icon'=>null], 'relatedModules' => ['Accounts'], 'moduleName' => 'BankAccounts'],
                    [107, 3147, 'company_name_a', 'u_yf_fcorectinginvoice_address', 1, 1, 'company_name_a', 'FL_COMPANY_NAME', 0, 0, '', '255', 12, 363, 1, 'V~O', 1, 0, 'BAS', 1, '', 0, '', null, 0, 0, 0, 0, '', '', 'type' => $importerType->stringType(),'blockLabel' => 'LBL_ADDRESS_INFORMATION', 'blockData' => ['label'=>'LBL_ADDRESS_INFORMATION','showtitle'=>0,'visible'=>0,'increateview'=>0,'ineditview'=>0,'indetailview'=>0,'display_status'=>2,'iscustom'=>0,'icon'=>null], 'moduleName' => 'FCorectingInvoice'],
                    [107, 3148, 'first_name_a', 'u_yf_fcorectinginvoice_address', 1, 1, 'first_name_a', 'First Name', 0, 0, '', '255', 13, 363, 1, 'V~O', 1, 0, 'BAS', 1, '', 0, '', null, 0, 0, 0, 0, '', '', 'type' => $importerType->stringType(),'blockLabel' => 'LBL_ADDRESS_INFORMATION', 'blockData' => ['label'=>'LBL_ADDRESS_INFORMATION','showtitle'=>0,'visible'=>0,'increateview'=>0,'ineditview'=>0,'indetailview'=>0,'display_status'=>2,'iscustom'=>0,'icon'=>null], 'moduleName' => 'FCorectingInvoice'],
                    [107, 3149, 'last_name_a', 'u_yf_fcorectinginvoice_address', 1, 1, 'last_name_a', 'Last Name', 0, 0, '', '255', 14, 363, 1, 'V~O', 1, 0, 'BAS', 1, '', 0, '', null, 0, 0, 0, 0, '', '', 'type' => $importerType->stringType(),'blockLabel' => 'LBL_ADDRESS_INFORMATION', 'blockData' => ['label'=>'LBL_ADDRESS_INFORMATION','showtitle'=>0,'visible'=>0,'increateview'=>0,'ineditview'=>0,'indetailview'=>0,'display_status'=>2,'iscustom'=>0,'icon'=>null], 'moduleName' => 'FCorectingInvoice'],
                    [107, 3150, 'vat_id_a', 'u_yf_fcorectinginvoice_address', 1, 1, 'vat_id_a', 'Vat ID', 0, 0, '', '50', 15, 363, 1, 'V~O', 1, 0, 'BAS', 1, '', 0, '', null, 0, 0, 0, 0, '', '', 'type' => $importerType->stringType(50),'blockLabel' => 'LBL_ADDRESS_INFORMATION', 'blockData' => ['label'=>'LBL_ADDRESS_INFORMATION','showtitle'=>0,'visible'=>0,'increateview'=>0,'ineditview'=>0,'indetailview'=>0,'display_status'=>2,'iscustom'=>0,'icon'=>null], 'moduleName' => 'FCorectingInvoice'],
                    [107, 3151, 'issue_time', 'u_yf_fcorectinginvoice', 1, 5, 'issue_time', 'FL_ISSUE_TIME', 0, 0, '', null, 16, 361, 1, 'D~M', 1, 0, 'BAS', 1, '', 0, '', null, 0, 0, 0, 0, '', '', 'type' => $importerType->date(),'blockLabel' => 'LBL_BASIC_DETAILS', 'blockData' => ['label'=>'LBL_BASIC_DETAILS','showtitle'=>0,'visible'=>0,'increateview'=>0,'ineditview'=>0,'indetailview'=>0,'display_status'=>2,'iscustom'=>0,'icon'=>null], 'moduleName' => 'FCorectingInvoice'],
                ];
                break;
            default:
                break;
        }

        return $fields;
    }
    private function preSchemaUpdateData()
    {
        $start = microtime(true);
        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s'));

        $db = \App\Db::getInstance();

        if(!$db->isTableExists('a_#__module_package')) {
            $db->createCommand('CREATE TABLE `a_yf_module_package` (
				`id` smallint(6) NOT NULL AUTO_INCREMENT,
				`name` varchar(50) NOT NULL,
				`label` varchar(50) NOT NULL,
				`description` varchar(50) NOT NULL,
				`icon` varchar(100) DEFAULT NULL,
				`sequence` smallint(6) DEFAULT NULL,
				`default_active` tinyint(1) DEFAULT 0,
				PRIMARY KEY (`id`)
			  ) ENGINE=InnoDB AUTO_INCREMENT=6 DEFAULT CHARSET=utf8mb4')->execute();
        }
        if(!$db->isTableExists('a_yf_module_reference')) {
            $db->createCommand('CREATE TABLE `a_yf_module_reference` (
				`id` smallint(5) unsigned NOT NULL AUTO_INCREMENT,
				`module_id` smallint(6) NOT NULL,
				`ref_table` varchar(50) NOT NULL,
				`ref_column` varchar(50) NOT NULL,
				`sort` tinyint(1) NOT NULL,
				`is_custom_field_table` tinyint(1) NOT NULL,
				PRIMARY KEY (`id`),
				KEY `fk_module_relations_entityname` (`module_id`),
				CONSTRAINT `fk_module_relations_entityname` FOREIGN KEY (`module_id`) REFERENCES `vtiger_entityname` (`tabid`) ON DELETE CASCADE
			  ) ENGINE=InnoDB AUTO_INCREMENT=279 DEFAULT CHARSET=utf8mb4')->execute();
        }
        if(!$db->isTableExists('a_yf_modules_packages')) {
            $db->createCommand('CREATE TABLE `a_yf_modules_packages` (
				`package_id` smallint(6) NOT NULL,
				`module_id` smallint(6) NOT NULL,
				KEY `a_yf_modules_packages_package_fk` (`package_id`),
				KEY `a_yf_modules_packages_module_fk` (`module_id`),
				CONSTRAINT `a_yf_modules_packages_module_fk` FOREIGN KEY (`module_id`) REFERENCES `vtiger_tab` (`tabid`) ON DELETE CASCADE,
				CONSTRAINT `a_yf_modules_packages_package_fk` FOREIGN KEY (`package_id`) REFERENCES `a_yf_module_package` (`id`) ON DELETE CASCADE
			  ) ENGINE=InnoDB DEFAULT CHARSET=utf8mb4;')->execute();
        }
        if(!$db->isTableExists('i_#__auth')) {
            $db->createCommand('CREATE TABLE `i_yf_auth` (
				`id` tinyint(3) unsigned NOT NULL AUTO_INCREMENT,
				`name` varchar(50) NOT NULL DEFAULT \'\',
				`providertype` varchar(20) NOT NULL,
				`isactive` tinyint(1) unsigned DEFAULT 0,
				`config` longtext CHARACTER SET utf8mb4 COLLATE utf8mb4_bin DEFAULT NULL CHECK (json_valid(`config`)),
				PRIMARY KEY (`id`),
				UNIQUE KEY `i_yf_auth_providertype_uidx` (`providertype`)
			  ) ENGINE=InnoDB DEFAULT CHARSET=utf8;')->execute();
        }
        if(!$db->isTableExists('u_yf_report_template_query')) {
            $db->createCommand('CREATE TABLE `u_yf_report_template_query` (
				`id` smallint(6) NOT NULL AUTO_INCREMENT,
				`parent_id` smallint(6) DEFAULT NULL,
				`base_module_id` smallint(6) NOT NULL,
				`data_sources` text NOT NULL,
				PRIMARY KEY (`id`),
				KEY `u_yf_report_template_query_parent_fk` (`parent_id`),
				KEY `u_yf_report_template_query_base_module_fk` (`base_module_id`),
				CONSTRAINT `u_yf_report_template_query_base_module_fk` FOREIGN KEY (`base_module_id`) REFERENCES `vtiger_entityname` (`tabid`) ON DELETE CASCADE,
				CONSTRAINT `u_yf_report_template_query_parent_fk` FOREIGN KEY (`parent_id`) REFERENCES `u_yf_report_template_query` (`id`) ON DELETE CASCADE
			  ) ENGINE=InnoDB DEFAULT CHARSET=utf8')->execute();
        }
        if(!$db->isTableExists('u_yf_report_template_query_expression')) {
            $db->createCommand('CREATE TABLE `u_yf_report_template_query_expression` (
				`id` smallint(6) NOT NULL AUTO_INCREMENT,
				`query_id` smallint(6) NOT NULL,
				`field` varchar(255) NOT NULL,
				`function` varchar(255) DEFAULT NULL,
				`group_by` tinyint(1) unsigned DEFAULT NULL,
				`order_by` varchar(4) DEFAULT NULL,
				`sort` tinyint(4) NOT NULL,
				PRIMARY KEY (`id`),
				KEY `u_yf_report_template_query_expression_query_fk` (`query_id`),
				CONSTRAINT `u_yf_report_template_query_expression_query_fk` FOREIGN KEY (`query_id`) REFERENCES `u_yf_report_template_query` (`id`) ON DELETE CASCADE
			  ) ENGINE=InnoDB DEFAULT CHARSET=utf8;')->execute();
        }
        if(!$db->isTableExists('u_yf_report_template_query_filter')) {
            $db->createCommand('CREATE TABLE `u_yf_report_template_query_filter` (
				`id` smallint(6) NOT NULL AUTO_INCREMENT,
				`filter_group_id` smallint(6) NOT NULL,
				`field` varchar(255) NOT NULL,
				`operator` varchar(25) NOT NULL,
				`value` varchar(255) DEFAULT NULL,
				`subquery_id` smallint(6) DEFAULT NULL,
				PRIMARY KEY (`id`),
				KEY `u_yf_report_template_query_filter_filter_group_fk` (`filter_group_id`),
				KEY `u_yf_report_template_query_filter_subquery_fk` (`subquery_id`),
				CONSTRAINT `u_yf_report_template_query_filter_filter_group_fk` FOREIGN KEY (`filter_group_id`) REFERENCES `u_yf_report_template_query_filter_group` (`id`) ON DELETE CASCADE,
				CONSTRAINT `u_yf_report_template_query_filter_subquery_fk` FOREIGN KEY (`subquery_id`) REFERENCES `u_yf_report_template_query` (`id`) ON DELETE SET NULL
			  ) ENGINE=InnoDB DEFAULT CHARSET=utf8;')->execute();
        }
        if(!$db->isTableExists('u_yf_report_template_query_filter_group')) {
            $db->createCommand('CREATE TABLE `u_yf_report_template_query_filter_group` (
				`id` smallint(6) NOT NULL AUTO_INCREMENT,
				`query_id` smallint(6) NOT NULL,
				`condition` varchar(3) NOT NULL,
				`parent_id` smallint(6) DEFAULT NULL,
				PRIMARY KEY (`id`),
				KEY `u_yf_report_template_query_filter_group_query_fk` (`query_id`),
				KEY `u_yf_report_template_query_filter_group_parent_fk` (`parent_id`),
				CONSTRAINT `u_yf_report_template_query_filter_group_parent_fk` FOREIGN KEY (`parent_id`) REFERENCES `u_yf_report_template_query_filter_group` (`id`) ON DELETE CASCADE,
				CONSTRAINT `u_yf_report_template_query_filter_group_query_fk` FOREIGN KEY (`query_id`) REFERENCES `u_yf_report_template_query` (`id`) ON DELETE CASCADE
			  ) ENGINE=InnoDB DEFAULT CHARSET=utf8;')->execute();
        }
        if(!$db->isTableExists('u_yf_report_template_query_join')) {
            $db->createCommand('CREATE TABLE `u_yf_report_template_query_join` (
				`id` smallint(6) NOT NULL AUTO_INCREMENT,
				`query_id` smallint(6) NOT NULL,
				`relation_id` smallint(5) unsigned NOT NULL,
				`type` varchar(5) NOT NULL,
				PRIMARY KEY (`id`),
				KEY `u_yf_report_template_query_join_query_fk` (`query_id`),
				KEY `u_yf_report_template_query_join_relation_fk` (`relation_id`),
				CONSTRAINT `u_yf_report_template_query_join_query_fk` FOREIGN KEY (`query_id`) REFERENCES `u_yf_report_template_query` (`id`) ON DELETE CASCADE,
				CONSTRAINT `u_yf_report_template_query_join_relation_fk` FOREIGN KEY (`relation_id`) REFERENCES `vtiger_relatedlists` (`relation_id`) ON DELETE CASCADE
			  ) ENGINE=InnoDB DEFAULT CHARSET=utf8;')->execute();
        }

        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s') . ' | ' . round((microtime(true) - $start) / 60, 2) . ' mim.');
    }

    private function migrateLdap()
    {
        $start = microtime(true);
        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s'));

        $db = \App\Db::getInstance();
        if($db->isTableExists('yetiforce_auth')) {
            $data = (new \App\Db\Query())->select(['param', 'value'])->from('yetiforce_auth')->where(['type' => 'ldap'])->createCommand()->queryAllByGroup();
            if($data) {
                $config = [
                    'server' => $data['server'] ?? '',
                    'port' => $data['port'] ?? '',
                    'domain' => $data['domain'] ?? '',
                ];
                $insertData['isactive'] = (int) ('true' === ($data['active'] ?? ''));
                if($insertData['isactive'] || 'testlab.local' !== $data['server']) {
                    $insertData['name'] = $config['server'];
                    $insertData['providertype'] = 'LDAP';
                    $insertData['config'] = \App\Json::encode($config);

                    $rows = [['i_yf_auth', $insertData]];
                    $this->log('  [INFO] Insert pre: ' . \App\Utils::varExport($rows));
                    $batchInsert = \App\Db\Updater::batchInsert($rows);
                    $this->log('  [INFO] Insert: ' . \App\Utils::varExport($batchInsert));
                }

                $db->createCommand()->delete('yetiforce_auth', ['type' => 'ldap'])->execute();
            }
            if(!(new \App\Db\Query())->from('yetiforce_auth')->exists()) {
                $this->importer->dropTable(['yetiforce_auth']);
            }
        }

        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s') . ' | ' . round((microtime(true) - $start) / 60, 2) . ' mim.');
    }

    private function updateData(): void
    {
        $start = microtime(true);
        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s'));
        $db = \App\Db::getInstance();
        $bankId = (new \App\Db\Query())->select(['id'])->from('yetiforce_currencyupdate_banks')->where(['bank_name' => 'NBR'])->scalar();

        $batchDelete = \App\Db\Updater::batchDelete([
            ['vtiger_settings_field', ['name' => 'LBL_VULNERABILITIES']],
            ['vtiger_settings_field', ['name' => 'Mail View']],
            ['vtiger_settings_field', ['name' => 'LBL_OSSMAIL']],
            ['vtiger_links', ['linkurl' => ['index.php?module=OSSMailView&view=ShowWidget&name=Graf','index.php?module=Leads&view=ShowWidget&name=LeadsByStatusConverted']]],
            ['yetiforce_currencyupdate_banks', ['bank_name' => 'NBR']],
            ['yetiforce_currencyupdate', ['bank_id' => (int) $bankId]],
            ['vtiger_rowheight', ['rowheight' => ['wide','narrow']]],
            ['yetiforce_mail_config', ['type' => 'autologin']],
        ]);
        $this->log('  [INFO] batchDelete: ' . \App\Utils::varExport($batchDelete));
        unset($batchDelete);

        $updates = [
            ['vtiger_field', ['displaytype' => 1], ['columnname' => 'projectid', 'tablename' => 'vtiger_projecttask', 'displaytype' => 9]],
            ['vtiger_eventhandlers', ['include_modules' => 'Accounts,Leads,Partners,Vendors,Competition,Contacts,Locations'], ['handler_class' => 'OpenStreetMap_OpenStreetMapHandler_Handler', 'event_name' => 'EntityAfterSave']],
            ['vtiger_settings_field', ['linkto' => 'index.php?module=Groups&parent=Settings&view=ListTable'], ['name' => 'USERGROUPLIST']],
            ['vtiger_links', ['handler_class' => 'Vtiger_CalendarActivitiesModel_Dashboard'], ['linkurl' => 'index.php?module=Home&view=ShowWidget&name=CalendarActivities']],
            ['vtiger_links', ['handler_class' => 'Vtiger_OverdueActivitiesModel_Dashboard'], ['linkurl' => 'index.php?module=Home&view=ShowWidget&name=OverdueActivities']],

            ['vtiger_settings_field', ['linkto' => 'index.php?module=OSSMailScanner&parent=Settings&view=List'], ['name' => 'Mail Scanner']],
            ['vtiger_settings_field', ['name'=>'LBL_AUTHENTICATION', 'iconpath'=>'yfi-ldap','description'=>'LBL_AUTHORIZATION_DESCRIPTION', 'linkto'=>'index.php?module=UserAuth&parent=Settings&view=List'], ['name'=>'LBL_AUTHORIZATION']],
            ['vtiger_cron_task', ['frequency'=>'21600'], ['handler_class'=>'OSSMailScanner_Verification_Cron']],
            ['vtiger_cron_task', ['frequency'=>'21600'], ['handler_class'=>'OSSMailScanner_Bind_Cron']],
            ['vtiger_cron_task', ['handler_class'=>'OSSMailScanner_Scann_Cron'], ['handler_class'=>'OSSMailScanner_Action_Cron']],
            ['vtiger_field', ['displaytype'=>'10'], ['tablename'=>'vtiger_users', 'columnname'=>'rowheight']],
            ['vtiger_field', ['fieldparams'=>'{"validate":["pwned","config"],"auto-generate":true,"strengthMeter":true}'], ['tablename'=>'vtiger_users', 'columnname'=>'user_password']],
            ['vtiger_field', ['fieldparams'=>'{"validate":["pwned","config"],"auto-generate":true,"strengthMeter":true,"copy":true}'], ['tablename'=>'u_yf_passwords', 'columnname'=>'password']],
            ['vtiger_users', ['rowheight'=>'medium']],
            ['vtiger_users', ['theme'=>'twilight']],
            ['vtiger_finvoice_status', ['presence'=>'0'], ['finvoice_status' => ['PLL_AWAITING_REALIZATION','PLL_CANCELLED','PLL_ACCEPTED']]],
        ];
        $batchUpdate = \App\Db\Updater::batchUpdate($updates);
        $this->log('  [INFO] batchUpdate: ' . \App\Utils::varExport($batchUpdate));
        unset($batchUpdate);

        $rows = [
            ['vtiger_eventhandlers', ['event_name' => 'EntityAfterSave', 'handler_class'=>'Vtiger_UpdateMapCoordinates_Handler', 'is_active'=>1,'include_modules'=>'Locations', 'exclude_modules'=>'','priority'=>5,'owner_id'=>0,'privileges'=>0], ['handler_class'=>'Vtiger_UpdateMapCoordinates_Handler']],
            ['a_yf_settings_modules', ['name' => 'WooCommerce', 'status'=>1, 'created_time'=> date('Y-m-d H:i:s')],['name' => 'WooCommerce']],
            ['a_yf_settings_modules', ['name' => 'MailServers', 'status'=>1, 'created_time'=> date('Y-m-d H:i:s')],['name' => 'MailServers']],
            ['a_yf_settings_modules', ['name' => 'MailSignature', 'status'=>1, 'created_time'=> date('Y-m-d H:i:s')],['name' => 'MailSignature']],
            ['a_yf_settings_modules', ['name' => 'Comarch', 'status'=>1, 'created_time'=> date('Y-m-d H:i:s')],['name' => 'Comarch']],
            ['a_yf_settings_modules', ['name' => 'UserAuth', 'status'=>1, 'created_time'=> date('Y-m-d H:i:s')],['name' => 'UserAuth']],

            ['vtiger_cron_task', ['status' => 1, 'name'=>'LBL_MAIL_SCANNER_NOTIFICATION', 'handler_class'=> 'OSSMailScanner_Notifications_Cron', 'frequency'=> 28800, 'sequence'=> 36, 'module' => 'OSSMailScanner'],['name'=>'LBL_MAIL_SCANNER_NOTIFICATION']],
            ['vtiger_cron_task', ['status' => 0, 'name'=>'LBL_WOOCOMMERCE', 'handler_class'=> 'Vtiger_WooCommerce_Cron', 'frequency'=> 300, 'sequence'=> 37],['name'=>'LBL_WOOCOMMERCE']],
            ['vtiger_eventhandlers', ['event_name' => 'EntityAfterSave', 'handler_class'=>'Products_UpdateModifiedTime_Handler', 'is_active'=>0,'include_modules'=>'Products', 'exclude_modules'=>'','priority'=>9,'owner_id'=>\App\Module::getModuleId('Products'),'privileges'=>0], ['handler_class'=>'Products_UpdateModifiedTime_Handler']],
            ['vtiger_settings_field', ['blockid' =>\vtlib\Deprecated::getSettingsBlockId('LBL_INTEGRATION'), 'name'=>'LBL_WOOCOMMERCE', 'iconpath'=>'fa-solid fa-cash-register','description'=>'LBL_WOOCOMMERCE_DESCRIPTION', 'linkto'=>'index.php?parent=Settings&module=WooCommerce&view=List','sequence'=>19,'active'=>0,'pinned'=>0, 'premium' => 1], ['name'=>'LBL_WOOCOMMERCE']],
            ['vtiger_settings_field', ['blockid' =>\vtlib\Deprecated::getSettingsBlockId('LBL_MAIL_TOOLS'), 'name'=>'LBL_MAIL_SIGNATURE', 'iconpath'=>'fa-solid fa-signature','description'=>'LBL_MAIL_SIGNATURE_DESCRIPTION', 'linkto'=>'index.php?parent=Settings&module=MailSignature&view=List','sequence'=>6,'active'=>0,'pinned'=>0, 'premium' => 0], ['name'=>'LBL_MAIL_SIGNATURE']],
            ['vtiger_settings_field', ['blockid' =>\vtlib\Deprecated::getSettingsBlockId('LBL_INTEGRATION'), 'name'=>'LBL_COMARCH', 'iconpath'=>'fa-brands fa-connectdevelop','description'=>'LBL_COMARCH_DESCRIPTION', 'linkto'=>'index.php?parent=Settings&module=Comarch&view=List','sequence'=>20,'active'=>0,'pinned'=>1, 'premium' => 1], ['name'=>'LBL_COMARCH']],

            ['vtiger_calendar_config', ['type' =>'reminder', 'name'=>'shared_persons', 'label'=>'LBL_RM_SHARING_ON','value'=>null], ['name'=>'shared_persons']],

            ['vtiger_eventhandlers', ['event_name' => 'EditViewPreSave', 'handler_class'=>'Reservations_DuplicateChecker_Handler', 'is_active'=>1,'include_modules'=>'Reservations', 'exclude_modules'=>'','priority'=>5,'owner_id'=>\App\Module::getModuleId('Reservations'),'privileges'=>1], ['handler_class'=>'Reservations_DuplicateChecker_Handler']],

            ['vtiger_eventhandlers', ['event_name' => 'EditViewPreSave', 'handler_class'=>'Calendar_RepeatEvents_Handler', 'is_active'=>1,'include_modules'=>'Calendar', 'exclude_modules'=>'','priority'=>5,'owner_id'=>\App\Module::getModuleId('Calendar'),'privileges'=>1], ['handler_class'=>'Calendar_RepeatEvents_Handler','event_name' => 'EditViewPreSave']],

            ['vtiger_eventhandlers', ['event_name' => 'PreDelete', 'handler_class'=>'Calendar_RepeatEvents_Handler', 'is_active'=>1,'include_modules'=>'Calendar', 'exclude_modules'=>'','priority'=>5,'owner_id'=>\App\Module::getModuleId('Calendar'),'privileges'=>1], ['event_name' => 'PreDelete','handler_class'=>'Calendar_RepeatEvents_Handler']],
            ['vtiger_eventhandlers', ['event_name' => 'PreStateChange', 'handler_class'=>'Calendar_RepeatEvents_Handler', 'is_active'=>1,'include_modules'=>'Calendar', 'exclude_modules'=>'','priority'=>5,'owner_id'=>\App\Module::getModuleId('Calendar'),'privileges'=>1], ['event_name' => 'PreStateChange','handler_class'=>'Calendar_RepeatEvents_Handler']],

            ['vtiger_cron_task', ['status' => 0, 'name'=>'LBL_COMARCH', 'handler_class'=> 'Vtiger_Comarch_Cron', 'frequency'=> 300, 'sequence'=> 38],['name'=>'LBL_COMARCH']],
            ['vtiger_cron_task', ['status' => 0, 'name'=>'LBL_DISPOSAL_LOGS_DATA', 'handler_class'=> 'Vtiger_DisposalLogsData_Cron', 'frequency'=> 86400, 'sequence'=> 39],['name'=>'LBL_DISPOSAL_LOGS_DATA']],

            ['vtiger_eventhandlers', ['event_name' => 'EditViewPreSave', 'handler_class'=>'Accounts_DuplicateShortName_Handler', 'is_active'=>1,'include_modules'=>'Accounts', 'exclude_modules'=>'','priority'=>5,'owner_id'=>\App\Module::getModuleId('Accounts'),'privileges'=>1], ['handler_class'=>'Accounts_DuplicateShortName_Handler']],

            ['vtiger_settings_field', ['blockid' =>\vtlib\Deprecated::getSettingsBlockId('LBL_MAIL_TOOLS'), 'name'=>'LBL_MAIL_SERVERS', 'iconpath'=>'fas fa-server','description'=>'LBL_MAIL_SERVERS_DESCRIPTION', 'linkto'=>'index.php?module=MailServers&parent=Settings&view=List','sequence'=>9,'active'=>0,'pinned'=>0, 'premium' => 0], ['name'=>'LBL_MAIL_SERVERS']],
        ];
        $batchInsert = \App\Db\Updater::batchInsert($rows);
        $this->log('  [INFO] Insert: ' . \App\Utils::varExport($batchInsert));


        // module tables data
        $rows = [];
        $dataReader = (new \App\Db\Query())->select(['name', 'tabid'])->from('vtiger_tab')->where(['or', ['isentitytype' => 1],['name' => 'Users']])->orderBy(['tabid'=> SORT_ASC])->createCommand()->query();
        while ($tabData = $dataReader->read()) {
            $tabId = $tabData['tabid'];
            $moduleName = $tabData['name'];

            if('ReportTemplate' === $moduleName) {
                $tables = [
                    'vtiger_crmentity'=>'crmid','u_yf_report_template'=>'id','u_yf_report_template_cf'=>'id'
                ];
                $customFieldTable = ['u_yf_report_template_cf'];
            } else {
                $focus = CRMEntity::getInstance($moduleName);
                $tables = $focus->tab_name_index;
                $customFieldTable = property_exists($focus, 'customFieldTable') ? $focus->customFieldTable : [];
            }


            $i = 1;
            foreach($tables as $tableName => $tableIndex) {
                if('u_yf_wapro_records_map' === $tableName || 'u_#__wapro_records_map' === $tableName || in_array($tableName, ['vtiger_activity_reminder','vtiger_senotesrel','vtiger_seproductsrel'])) {
                    continue;
                }
                $customTable = in_array($tableName, $customFieldTable);
                $rows[] = ['a_yf_module_reference', ['module_id' => $tabId, 'ref_table'=>$tableName, 'ref_column'=>$tableIndex,'sort'=>$i++, 'is_custom_field_table'=>(int) $customTable], ['module_id' => $tabId, 'ref_table'=>$tableName, 'ref_column'=>$tableIndex]];
            }
        }
        $dataReader->close();

        $batchInsert = \App\Db\Updater::batchInsert($rows);
        $this->log('  [INFO] Insert: ' . \App\Utils::varExport($batchInsert));


        // user internal_mailer
        $updates = [];
        $dataReader = (new \App\Db\Query())->select(['id', 'internal_mailer'])->from('vtiger_users')->createCommand()->query();
        while ($tabData = $dataReader->read()) {
            $id = $tabData['id'];
            $internalMailer = $tabData['internal_mailer'];
            if((string) $internalMailer === '1' && file_exists(ROOT_DIRECTORY . '/public_html/modules/OSSMail/roundcube/')) {
                $internalMailer = 'InternalClient';
            } elseif (empty($internalMailer) || '0' === $internalMailer) {
                $internalMailer = 'Base';
            }
            $updates[] = ['vtiger_users', ['internal_mailer' => $internalMailer], ['id' => $id]];
        }
        $dataReader->close();

        $updates[] = ['vtiger_field', ['uitype' => 87, 'fieldlabel' => 'FL_INTERNAL_MAIL_COMPOSER'], ['columnname' => 'internal_mailer', 'tablename' => 'vtiger_users']];
        $batchUpdate = \App\Db\Updater::batchUpdate($updates);
        $this->log('  [INFO] batchUpdate: ' . \App\Utils::varExport($batchUpdate));
        unset($batchUpdate);

        // callhistory color
        $colors = (new \App\Db\Query())->select(['color'])->from('vtiger_callhistorytype')->column();
        if(!array_filter($colors)) {
            $updates[] = ['vtiger_callhistorytype', ['color' => "FFF176"], ['callhistorytype' => "Outgoing missed"]];
            $updates[] = ['vtiger_callhistorytype', ['color' => "81C784"], ['callhistorytype' => "Outgoing received"]];
            $updates[] = ['vtiger_callhistorytype', ['color' => "4DD0E1"], ['callhistorytype' => "Incoming"]];
            $updates[] = ['vtiger_callhistorytype', ['color' => "F44336"], ['callhistorytype' => "Missed"]];
            $updates[] = ['vtiger_callhistorytype', ['color' => "90A4AE"], ['callhistorytype' => "Voicemail"]];
            $batchUpdate = \App\Db\Updater::batchUpdate($updates);
            $this->log('  [INFO] batchUpdate (callhistory color): ' . \App\Utils::varExport($batchUpdate));
        }


        // maps
        $batchDelete = \App\Db\Updater::batchDelete([
            ['s_yf_address_finder_config', ['type' => 'YetiForceGeocoder']],
            ['s_yf_address_finder_config', ['val' => 'YetiForceGeocoder']],
        ]);
        $this->log('  [INFO] batchDelete (YetiForceGeocoder): ' . \App\Utils::varExport($batchDelete));
        unset($batchDelete);

        // mail messages
        $moduleModel = \Vtiger_Module_Model::getInstance('OSSMailView');
        foreach(['uid'=>'msgid', 'mid'=>'uid'] as $oldName => $newName) {
            if ($fieldModel = $moduleModel->getFieldByName($oldName)) {
                try {
                    $this->renameField($fieldModel, $newName, ['columnname' => $newName]);
                } catch (\Throwable $e) {
                    $message = '[ERROR] ' . __METHOD__ . ": {$moduleName}:{$oldName} " . $e->__toString();
                    $this->log($message);
                    \App\Log::error($message);
                }
            } else {
                $this->log("[Warning] Skip renaming {$moduleName}:{$oldName}, field not exists");
            }
        }


        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s') . ' | ' . round((microtime(true) - $start) / 60, 2) . ' mim.');
    }

    private function renameField($fieldModel, string $newName, array $updateData = [])
    {
        $start = microtime(true);
        $this->log(__METHOD__ . " {$newName} | " . date('Y-m-d H:i:s'));

        $db = \App\Db::getInstance();
        $transaction = $db->beginTransaction();
        try {
            $dbCommand = $db->createCommand();

            $fldModule = $fieldModel->getModuleName();
            $fieldname = $fieldModel->getName();
            $tabId = $fieldModel->getModuleId();

            $updateData['fieldname'] = $newName;

            $dbCommand->update('vtiger_field', $updateData, ['fieldname' => $fieldname, 'tabid' => $tabId])->execute();
            $dbCommand->update('vtiger_cvcolumnlist', ['field_name' => $newName], ['field_name' => $fieldname, 'module_name' => $fldModule])->execute();
            $dbCommand->update('u_#__cv_condition', ['field_name' => $newName], ['field_name' => $fieldname, 'module_name' => $fldModule])->execute();

            if ('picklist' === $fieldModel->getFieldDataType() || 'multipicklist' === $fieldModel->getFieldDataType()) {
                $tableName = 'vtiger_' . $fieldname;
                $newTableName = 'vtiger_' . $newName;
                if ($db->isTableExists($newTableName) && $db->isTableExists($tableName)) {
                    $currentValues = (new \App\Db\Query())->select([$fieldname])->from($tableName)->column();
                    $newPicklistValue = (new \App\Db\Query())->select([$newName])->from($newTableName)->column();
                    $sortId = \count($newPicklistValue);
                    foreach (array_diff($currentValues, $newPicklistValue) as $value) {
                        $dbCommand->insert($newTableName, [$newName => $value, 'presence' => 1, 'sortorderid' => ++$sortId])->execute();
                        $this->log("[INFO] Add value to: {$newTableName}:{$value};");
                    }
                }
                $query = (new \App\Db\Query())->from('vtiger_field')
                    ->where(['fieldname' => $fieldname])
                    ->andWhere(['in', 'uitype', [15, 16, 33]]);
                $dataReader = $query->createCommand()->query();
                if (!$dataReader->count() && $db->isTableExists($tableName)) {
                    if (15 === $fieldModel->getUIType() || 33 === $fieldModel->getUIType()) {
                        $picklistId = (new \App\Db\Query())->select(['picklistid'])->from('vtiger_picklist')->where(['name' => $fieldname])->scalar();
                        $newPicklistId = (new \App\Db\Query())->select(['picklistid'])->from('vtiger_picklist')->where(['name' => $newName])->scalar();
                        if ($picklistId && $newPicklistId) {
                            $dbCommand->delete('vtiger_picklist', ['name' => $fieldname])->execute();
                            $dbCommand->delete('vtiger_role2picklist', ['picklistid' => $picklistId])->execute();
                        } elseif ($picklistId && !$newPicklistId) {
                            $dbCommand->update('vtiger_picklist', ['name' => $newName], ['name' => $fieldname])->execute();
                        }
                    }
                    $dbCommand->dropTable($tableName)->execute();
                    if ($db->isTableExists($tableName . '_seq')) {
                        $dbCommand->dropTable($tableName . '_seq')->execute();
                    }
                    $dbCommand->delete('vtiger_picklist', ['name' => $fieldname])->execute();
                }
                $dbCommand->update('vtiger_picklist_dependency', ['sourcefield' => $newName], ['tabid' => $tabId, 'sourcefield' => $fieldname])->execute();
                $dbCommand->update('vtiger_picklist_dependency', ['targetfield' => $newName], ['tabid' => $tabId, 'targetfield' => $fieldname])->execute();
            }
            $entityFieldInfo = \App\Module::getEntityInfo($fldModule);
            $fieldsName = $entityFieldInfo['fieldnameArr'];
            $searchColumns = $entityFieldInfo['searchcolumnArr'];
            if (\in_array($fieldname, $fieldsName) || \in_array($fieldname, $searchColumns)) {
                if (false !== ($key = array_search($fieldname, $fieldsName))) {
                    $fieldsName[$key] = $newName;
                }
                if (false !== ($key = array_search($fieldname, $searchColumns))) {
                    $searchColumns[$key] = $newName;
                }
                $dbCommand->update(
                    'vtiger_entityname',
                    ['fieldname' => implode(',', $fieldsName), 'searchcolumn' => implode(',', $searchColumns)],
                    ['modulename' => $entityFieldInfo['modulename']]
                )->execute();
                \App\Cache::delete('ModuleEntityById', $tabId);
                \App\Cache::delete('ModuleEntityByName', $fldModule);
            }
            $db->createCommand("UPDATE com_vtiger_workflowtasks SET task = REPLACE(task, '{$fieldname}', '{$newName}') WHERE task LIKE '{$fieldname}'")->execute();
            $db->createCommand("UPDATE com_vtiger_workflows SET test = REPLACE(test, '{$fieldname}', '{$newName}') WHERE test LIKE '{$fieldname}'")->execute();
            $transaction->commit();
        } catch (\Throwable $ex) {
            $transaction->rollBack();
            throw $ex;
        }
        \App\Cache::delete('ModuleFields', $fieldModel->getModuleId());
        \App\Cache::staticDelete('ModuleFields', $fieldModel->getModuleId());

        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s') . ' | ' . round((microtime(true) - $start) / 60, 2) . ' mim.');
    }

    private function updateSettingsMenu()
    {
        $start = microtime(true);
        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s'));
        $db = \App\Db::getInstance();

        $batchDelete = \App\Db\Updater::batchDelete([
            ['vtiger_settings_field', ['linkto' => 'index.php?parent=Settings&module=MailRbl&view=Index']],
            ['a_yf_settings_modules', ['name' => 'MailRbl']],
        ]);
        $this->log('  [INFO] batchDelete: ' . \App\Utils::varExport($batchDelete));

        $db->createCommand("UPDATE
		`vtiger_settings_field`
	  SET
		sequence =
		CASE
		  WHEN `name` = 'LBL_MAIL_SMTP'
		  THEN 1
		  WHEN `name` = 'LBL_MAIL_SERVERS'
		  THEN 2
		  WHEN `name` = 'Mail Scanner'
		  THEN 3
		  WHEN `name` = 'LBL_AUTOLOGIN'
		  THEN 4
		  WHEN `name` = 'LBL_MAIL_GENERAL_CONFIGURATION'
		  THEN 5
		  WHEN `name` = 'LBL_MAIL_SIGNATURE'
		  THEN 6
		  WHEN `name` = 'LBL_EMAILS_TO_SEND'
		  THEN 7
		END
	  WHERE `name` IN (
		  'LBL_MAIL_SMTP',
		  'LBL_MAIL_SERVERS',
		  'Mail Scanner',
		  'LBL_AUTOLOGIN',
		  'LBL_MAIL_SIGNATURE',
		  'LBL_MAIL_GENERAL_CONFIGURATION',
		  'LBL_EMAILS_TO_SEND'
		);")->execute();

        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s') . ' | ' . round((microtime(true) - $start) / 60, 2) . ' mim.');
    }

    private function migrateMap()
    {
        $start = microtime(true);
        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s'));

        $defaultCor = \App\Config::module('OpenStreetMap', 'coordinatesServer') ?: 'Nominatim';
        $defaultRout = \App\Config::module('OpenStreetMap', 'routingServer') ?: 'Osrm';
        $defaultLayer = \App\Config::module('OpenStreetMap', 'tileLayerServer') ?: 'OpenStreetMapDefault';
        $defaultLayer = str_replace(' ', '', $defaultLayer);
        $grapherKey = \App\Config::module('OpenStreetMap', 'routingServers')['GraphHopper']['params']['key'] ?? '';
        if($grapherKey && 'b16b1d60-3c8c-4cd6-bae6-07493f23e589' !== $grapherKey) {
            $grapherKey = \App\Encryption::getInstance()->encrypt($grapherKey);
        } else {
            $grapherKey = '';
        }
        $rows = [
            ['a_yf_map_coordinates_driver_config', ['driver' => 'GoogleGeocoding'], ['driver' => 'GoogleGeocoding']],
            ['a_yf_map_coordinates_driver_config', ['driver' => 'Nominatim', 'active' => (int) ($defaultCor === 'Nominatim')], ['driver' => 'Nominatim']],
            ['a_yf_map_coordinates_driver_config', ['driver' => 'OpenCage', 'active' =>  (int)($defaultCor === 'OpenCage')], ['driver' => 'OpenCage']],
            ['a_yf_map_coordinates_driver_config', ['driver' => 'GraphHopper', 'active' =>  (int)($defaultCor === 'GraphHopper')], ['driver' => 'GraphHopper']],

            ['a_yf_map_routing_driver_config', ['driver' => 'GoogleRoutes', 'params'=> '{"travelMode":"DRIVE","routingPreference":"TRAFFIC_UNAWARE"}' ], ['driver' => 'GoogleRoutes']],
            ['a_yf_map_routing_driver_config', ['driver' => 'Osrm', 'active' => (int) ($defaultRout === 'Osrm')], ['driver' => 'Osrm']],
            ['a_yf_map_routing_driver_config', ['driver' => 'GraphHopper', 'active' => (int) ($defaultRout === 'GraphHopper'),'api_key' => $grapherKey , 'params'=> '{"profile":"car"}' ], ['driver' => 'GraphHopper']],

            ['a_yf_map_tile_layer_driver_config', ['driver' => 'OpenStreetMapDefault','active' => (int) ($defaultLayer === 'OpenStreetMapDefault')], ['driver' => 'OpenStreetMapDefault']],
            ['a_yf_map_tile_layer_driver_config', ['driver' => 'OpenStreetMapHOT','active' => (int) ($defaultLayer === 'OpenStreetMapHOT')], ['driver' => 'OpenStreetMapHOT']],
            ['a_yf_map_tile_layer_driver_config', ['driver' => 'EsriWorldTopoMap','active' => (int) ($defaultLayer === 'EsriWorldTopoMap')], ['driver' => 'EsriWorldTopoMap']],
            ['a_yf_map_tile_layer_driver_config', ['driver' => 'EsriWorldImagery','active' => (int) ($defaultLayer === 'EsriWorldImagery')], ['driver' => 'EsriWorldImagery']],
            ['a_yf_map_tile_layer_driver_config', ['driver' => 'GoogleMaps','active' => (int) ($defaultLayer === 'GoogleMaps')], ['driver' => 'GoogleMaps']],

        ];
        $batchInsert = \App\Db\Updater::batchInsert($rows);
        $this->log('  [INFO] Insert: ' . \App\Utils::varExport($batchInsert));

        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s') . ' | ' . round((microtime(true) - $start) / 60, 2) . ' mim.');
    }
    private function inventoryDefaults()
    {
        $start = microtime(true);
        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s'));

        $updates = [];
        $query = (new \App\Db\Query())->select(['name'])->from('vtiger_tab')->where(['type' => 1]);
        $dataReader = $query->createCommand()->query();
        while ($moduleName = $dataReader->readColumn(0)) {
            $invModel = Vtiger_Inventory_Model::getInstance($moduleName);
            $table = $invModel->getTableName();
            $updates[] = [$table, ['defaultvalue' => ''], ['invtype' => "DiscountMode", 'defaultvalue' => '0']];
            $updates[] = [$table, ['defaultvalue' => ''], ['invtype' => "TaxMode", 'defaultvalue' => '0']];
        }
        $dataReader->close();

        $batchUpdate = \App\Db\Updater::batchUpdate($updates);
        $this->log('  [INFO] batchUpdate: ' . \App\Utils::varExport($batchUpdate));

        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s') . ' | ' . round((microtime(true) - $start) / 60, 2) . ' mim.');
    }

    public function modulePackage()
    {
        $start = microtime(true);
        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s'));

        $data = [
            ['1','Sales','LBL_MODULE_PACKAGE_SALES','LBL_MODULE_PACKAGE_SALES_DESCRIPTION','fa-coins','0','1'],
            ['2','Accounting','LBL_MODULE_PACKAGE_ACCOUNTING','LBL_MODULE_PACKAGE_ACCOUNTING_DESCRIPTION','fa-file-invoice-dollar','1','0'],
            ['3','ProjectManagement','LBL_MODULE_PACKAGE_PROJECT_MANAGEMENT','LBL_MODULE_PACKAGE_PROJECT_MANAGEMENT_DESCRIPTION','fa-list-check','2','0'],
            ['4','Marketing','LBL_MODULE_PACKAGE_MARKETING','LBL_MODULE_PACKAGE_MARKETING_DESCRIPTION','fa-ranking-star','3','0'],
            ['5','Support','LBL_MODULE_PACKAGE_SUPPORT','LBL_MODULE_PACKAGE_SUPPORT_DESCRIPTION','fa-suitcase-medical','4','0']
        ];
        $rows = [];
        foreach ($data as $value) {
            $rows[] = ['a_yf_module_package', ['id' => $value[0], 'name'=>$value[1], 'label' => $value[2],'description' => $value[3],'icon' =>$value[4],'sequence' => $value[5],'default_active' => $value[6]], ['name' => $value[1]]];
        }
        $batchInsert = \App\Db\Updater::batchInsert($rows);
        $this->log('  [INFO] Insert: ' . \App\Utils::varExport($batchInsert));


        $data =[
            ['1','Leads'],
            ['1','PriceBooks'],
            ['1','SCalculations'],
            ['1','SQuotes'],
            ['1','SRecurringOrders'],
            ['1','SRequirementsCards'],
            ['1','SSalesProcesses'],
            ['1','SQuoteEnquiries'],
            ['1','SSingleOrders'],
            ['2','FBookkeeping'],
            ['2','FCorectingInvoice'],
            ['2','FInvoice'],
            ['2','FInvoiceCost'],
            ['2','FInvoiceProforma'],
            ['2','PaymentsIn'],
            ['2','PaymentsOut'],
            ['5','Faq'],
            ['5','HelpDesk'],
            ['5','KnowledgeBase'],
            ['5','ServiceContracts'],
            ['3','Project'],
            ['3','ProjectMilestone'],
            ['3','ProjectTask'],
            ['4','Leads'],
            ['4','Campaigns'],
            ['4','Occurrences'],
            ['4','Locations']
        ];
        $rows = [];
        foreach ($data as $value) {
            $tabId = \App\Module::getModuleId($value[1]);
            $rows[] = ['a_yf_modules_packages', ['package_id' => $value[0], 'module_id'=>$tabId], ['package_id' => $value[0], 'module_id'=>$tabId]];
        }
        $batchInsert = \App\Db\Updater::batchInsert($rows);
        $this->log('  [INFO] Insert: ' . \App\Utils::varExport($batchInsert));

        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s') . ' | ' . round((microtime(true) - $start) / 60, 2) . ' mim.');
    }
    private function changeUitypeForPicklist()
    {
        $start = microtime(true);
        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s'));
        $db = \App\Db::getInstance();

        $schema = $db->getSchema();
        $dbCommand = $db->createCommand();
        $roleIds = (new \App\Db\Query())->select(['roleid'])->from('vtiger_role')->column();
        $query = (new \App\Db\Query())->from('vtiger_field')
            ->where(['uitype' => 16])
            ->andWhere(['fieldname' => 'svendorenquiries_status']);

        $dataReader = $query->createCommand()->query();
        while ($row = $dataReader->read()) {
            $picklistTable = 'vtiger_' . $row['fieldname'];
            $tableSchema = $schema->getTableSchema($picklistTable);
            if(!isset($tableSchema->columns['picklist_valueid'])) {
                $dbCommand->addColumn($picklistTable, 'picklist_valueid', $schema->createColumnSchemaBuilder(\yii\db\Schema::TYPE_INTEGER, 10)->notNull()->defaultValue(0))->execute();
            }

            $dbCommand->insert('vtiger_picklist', ['name' => $row['fieldname']])->execute();
            $newPicklistId = (new \App\Db\Query())->select(['picklistid'])->from('vtiger_picklist')->where(['name' => $row['fieldname']])->scalar();
            if (!$newPicklistId) {
                $newPicklistId = $db->getLastInsertID('vtiger_picklist_picklistid_seq');
            }
            $identifier = $row['fieldname'] . 'id';
            $query2 = (new \App\Db\Query())->select([$identifier, 'sortorderid'])->from($picklistTable);
            $dataReader2 = $query2->createCommand()->query();
            while ($picklistRow = $dataReader2->read()) {
                $newPicklistValueId = $db->getUniqueID('vtiger_picklistvalues');
                $dbCommand->update($picklistTable, ['picklist_valueid' => $newPicklistValueId], [$identifier => $picklistRow[$identifier]])->execute();
                $insertedData = [];
                foreach ($roleIds as $value) {
                    $insertedData[] = [$value, $newPicklistValueId, $newPicklistId, $picklistRow['sortorderid']];
                }
                $dbCommand->batchInsert('vtiger_role2picklist', ['roleid', 'picklistvalueid', 'picklistid', 'sortid'], $insertedData)->execute();
            }
            $dataReader2->close();
            $dbCommand->update('vtiger_field', ['uitype' => 15], ['fieldid' => $row['fieldid']])->execute();
        }
        $dataReader->close();

        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s') . ' | ' . round((microtime(true) - $start) / 60, 2) . ' mim.');
    }

    public function migrateMailConfig()
    {
        $start = microtime(true);
        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s'));
        $db = \App\Db::getInstance();

        if(!$db->isTableExists('vtiger_ossmailscanner_config')) {
            $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s') . ' | ' . round((microtime(true) - $start) / 60, 2) . ' mim.');
            return;
        }

        $open = '_self' !== (new \App\Db\Query())->select(['value'])->from('vtiger_ossmailscanner_config')->where(['conf_type'=>'email_list','parameter'=>'target'])->scalar();
        \App\Db\Updater::batchUpdate([['vtiger_users', ['mail_popup' => (int) $open]]]);


        $query = (new \App\Db\Query())->from('vtiger_ossmailscanner_config');
        $query->orderBy(['parameter' => SORT_DESC]);
        $dataReader = $query->createCommand()->query();
        $oldConfig = [];
        while ($row = $dataReader->read()) {
            $return[$row['conf_type']][$row['parameter']] = $row['value'];
        }
        $dataReader->close();

        $emailException = [];
        $domainException = [];
        $old = $oldConfig['exceptions']['crating_mails'] ?? '';
        $old = $old ? explode(',', $old) : [];

        foreach ($old as $value) {
            if(!$value) {
                continue;
            }
            if(false !== strpos($value, '@')) {
                $domainException[] = substr($value, 1);
            } else {
                $emailException[] = ["e"=>$value,"o"=>0];
            }
        }
        $mailConfig = (new \App\Db\Query())->select(['name','value'])->from('yetiforce_mail_config')->createCommand()->queryAllByGroup();
        $updateData = [
            ['type' => 'mailIcon','name' => 'showNumberUnreadEmails','value' => $mailConfig['showNumberUnreadEmails'] === 'true' ? '1' : '0','sequence' => 2],
            ['type' => 'mailIcon','name' => 'showMailIcon','value' => $mailConfig['showMailIcon'] === 'true' ? '1' : '0','sequence' => 1],
            ['type' => 'mailIcon','name' => 'timeCheckingMail', 'value' => $mailConfig['timeCheckingMail'] ?: '60','sequence' => 3],
            // ['type' => 'signature','name' => 'signature','value' => $mailConfig['signature'],'sequence' => 2],
            // ['type' => 'signature','name' => 'addSignature','value' => $mailConfig['signature'] === 'true' ? '1' : '0','sequence' => 1],
            ['type' => 'scanner','name' => 'email_for_notification','value' => $oldConfig['cron']['email'] ?? '','sequence' => 1],
            ['type' => 'scanner','name' => 'time_for_notification','value' => $oldConfig['cron']['time'] ?? '0','sequence' => 2],
            ['type' => 'scanner','name' => 'limit','value' => \App\Config::performance('NUMBERS_EMAILS_DOWNLOADED_DURING_ONE_SCANNING', 100),'sequence' => 3],
            ['type' => 'scanner','name' => 'deactivation_time','value' => '48','sequence' => 4],
            ['type' => 'scanner','name' => 'flag_seen','value' => '0','sequence' => 2],
            ['type' => 'scanner','name' => 'email_exceptions','value' => $emailException ? \App\Json::encode($emailException) : '','sequence' => 6],
            ['type' => 'scanner','name' => 'domain_exceptions','value' => $domainException ? implode(',', $domainException) : '','sequence' => 7],
        ];
        foreach ($updateData as $config) {
            $exists = (new \App\Db\Query())->select(['name','value'])->from('yetiforce_mail_config')->where(['name'=>$config['name']])->exists();
            if($exists) {
                $db->createCommand()->update('yetiforce_mail_config', $config, ['name'=>$config['name']])->execute();
            } else {
                $db->createCommand()->insert('yetiforce_mail_config', $config)->execute();
            }
        }
        $db->createCommand()->delete('yetiforce_mail_config', ['type' => 'signature'])->execute();

        $addSign = $mailConfig['signature'] === 'true';
        $sign = $mailConfig['signature'];
        if($sign) {
            $db->createCommand()->insert('s_yf_mail_signature', ['name' => 'Blank', 'status' => (int) $addSign, 'default' => (int) $addSign, 'body' => $sign])->execute();
        }


        $db->createCommand("UPDATE
		`vtiger_settings_field`
	  SET
		sequence =
		CASE
		  WHEN `name` = 'LBL_MAIL_SMTP'
		  THEN 1
		  WHEN `name` = 'LBL_MAIL_SERVERS'
		  THEN 2
		  WHEN `name` = 'Mail Scanner'
		  THEN 3
		  WHEN `name` = 'LBL_AUTOLOGIN'
		  THEN 4
		  WHEN `name` = 'LBL_MAIL_GENERAL_CONFIGURATION'
		  THEN 5
		  WHEN `name` = 'LBL_MAIL_SIGNATURE'
		  THEN 6
		  WHEN `name` = 'LBL_EMAILS_TO_SEND'
		  THEN 7
		END
	  WHERE `name` IN (
		  'LBL_MAIL_SMTP',
		  'LBL_MAIL_SERVERS',
		  'Mail Scanner',
		  'LBL_AUTOLOGIN',
		  'LBL_MAIL_SIGNATURE',
		  'LBL_MAIL_GENERAL_CONFIGURATION',
		  'LBL_EMAILS_TO_SEND'
		);")->execute();

        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s') . ' | ' . round((microtime(true) - $start) / 60, 2) . ' mim.');
    }

    public function migrateMailServer()
    {
        $start = microtime(true);
        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s'));
        $db = \App\Db::getInstance();

        $smtp_host = \App\Config::module('OSSMail', 'smtp_host', 'ssl://smtp.gmail.com:465');
        $imap_host = \App\Config::module('OSSMail', 'imap_host');
        $exists = (new \App\Db\Query())->from('s_yf_mail_servers')->exists();
        if(!$imap_host || $exists || !\App\Module::isModuleActive('OSSMail') || !file_exists(ROOT_DIRECTORY . '/public_html/modules/OSSMail/roundcube/') || 'ssl://smtp.gmail.com:465' === $smtp_host) {
            $this->log('  [INFO] Skip');
            $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s') . ' | ' . round((microtime(true) - $start) / 60, 2) . ' mim.');
            return;
        }

        $username_domain = \App\Config::module('OSSMail', 'username_domain');
        $imap_host = \App\Config::module('OSSMail', 'imap_host');
        $validate_cert = (int) \App\Config::module('OSSMail', 'validate_cert');
        $identities_level = (int) \App\Config::module('OSSMail', 'identities_level');



        $parseHost = parse_url($smtp_host);
        $hostSmtp = $parseHost['host']??'';
        $sslModeSmtp = (isset($parseHost['scheme']) && \in_array($parseHost['scheme'], ['ssl', 'tls'])) ? $parseHost['scheme'] : 'tls';
        $portSmtp = $parseHost['port'] ?? 587;

        $imap_hosts = is_array($imap_host) ? $imap_host : [$imap_host];
        $i = 0;
        foreach ($imap_hosts as $imap_host) {

            $parseHost = parse_url($imap_host);

            $hostImap = $parseHost['host']??'';
            $sslModeImap = (isset($parseHost['scheme']) && \in_array($parseHost['scheme'], ['ssl', 'tls'])) ? $parseHost['scheme'] : 'tls';
            $portImap = $parseHost['port'] ?? 143;

            $db->createCommand()->insert('s_yf_mail_servers', [
                'name' => 'Mail server ' . ++$i, 'auth_method' => 'basic', 'status' => 1, 'visible' => 0,
                'validate_cert' => $validate_cert, 'imap_encrypt' => $sslModeImap, 'imap_host' => $hostImap, 'imap_port' => $portImap,
                'smtp_encrypt' => $sslModeSmtp, 'smtp_host' => $hostSmtp, 'smtp_port' => $portSmtp, 'identities_level' => $identities_level,
            ])->execute();
        }



        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s') . ' | ' . round((microtime(true) - $start) / 60, 2) . ' mim.');
    }

    public function migrateMailAccounts()
    {
        $start = microtime(true);
        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s'));

        $db = \App\Db::getInstance();
        $db->createCommand('TRUNCATE TABLE u_yf_mail_quantities;')->execute();

        $servers = (new \App\Db\Query())->select(['imap_host','id'])->from('s_yf_mail_servers')->createCommand()->queryAllByGroup();
        if(!$servers) {
            $this->log('  [INFO] Skip: maile server not exists');
            $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s') . ' | ' . round((microtime(true) - $start) / 60, 2) . ' mim.');
        }

        $newScannerFields = [];
        $oldScannFields = (new \App\Db\Query())->select(['value'])->from('vtiger_ossmailscanner_config')->where(['conf_type'=>'emailsearch','parameter'=>'fields'])->scalar() ?: '';
        $oldScannFields = $oldScannFields ? explode(',', $oldScannFields) : [];
        foreach ($oldScannFields as $value) {
            [$fieldName, $moduleName, $uiType] = explode('=', $value);
            $fieldId = (new \App\Db\Query())->select(['fieldid'])->from('vtiger_field')->where(['tabid'=>\App\Module::getModuleId($moduleName),'fieldname'=>$fieldName])->scalar();
            if($fieldId) {
                $newScannerFields[] = "{$fieldId}|{$moduleName}|{$fieldName}|{$uiType}";
            }
        }


        // accounts
        $dataReader = (new \App\Db\Query())->from('roundcube_users')->createCommand()->query();
        while ($rData = $dataReader->read()) {
            $newScannerFieldsPerUser = $newScannerFields;
            $rcId = $rData['user_id'];
            $login = $rData['username'];
            $host = $rData['mail_host'];
            $crm_user_id = $rData['crm_user_id'];
            $pwd = $rData['password'];
            $oldStatus = (int) $rData['crm_status'];
            $oldActions = (string) $rData['actions'];
            $serverId = $servers[$host] ?? null;
            if(!$serverId) {
                $this->log("  [INFO] Skip ($login): maile server not exists for host: {$host}");
                continue;
            }
            $private = (int) (\App\User::isExists($crm_user_id) && $login === \App\User::getUserModel($crm_user_id)->getDetail('email1'));
            $statuses = [0 => 'PLL_ACTIVE', 1 => 'PLL_LOCKED', 2 => 'PLL_INACTIVE', 4 => 'PLL_INACTIVE', 3 => 'PLL_INACTIVE'];
            $status = $statuses[$oldStatus] ?? 'PLL_INACTIVE';

            $oldActions = explode(',', $oldActions);
            $actions = [];
            foreach ($oldActions as $oldAction) {
                switch ($oldAction) {
                    case 'CreatedEmail':
                        $actions[] = 'CreatedMail';
                        break;
                    case 'CreatedHelpDesk':
                        $actions[] = 'CreatedHelpDesk';
                        break;
                    case 'BindHelpDesk':
                        $fieldAlias = $this->getAliasForScanner('HelpDesk', 'ticket_no');
                        if($fieldAlias) {
                            $newScannerFieldsPerUser[] = $fieldAlias;
                        }
                        $newBindAction = 'BindMail';
                        if(!in_array($newBindAction, $actions)) {
                            $actions[] = $newBindAction;
                        }
                        if(!empty(\Config\Modules\OSSMailScanner::$helpdeskBindNextWaitForResponseStatus)) {
                            $newBindAction = 'OpenHelpDesk';
                            $actions[] = $newBindAction;
                        }
                        break;
                    case 'BindCampaigns':
                        $fieldAlias = $this->getAliasForScanner('Campaigns', 'campaign_no');
                        if($fieldAlias) {
                            $newScannerFieldsPerUser[] = $fieldAlias;
                        }
                        $newBindAction = 'BindMail';
                        if(!in_array($newBindAction, $actions)) {
                            $actions[] = $newBindAction;
                        }
                        break;
                    case 'BindProject':
                        $fieldAlias = $this->getAliasForScanner('Project', 'project_no');
                        if($fieldAlias) {
                            $newScannerFieldsPerUser[] = $fieldAlias;
                        }
                        $newBindAction = 'BindMail';
                        if(!in_array($newBindAction, $actions)) {
                            $actions[] = $newBindAction;
                        }
                        break;
                    case 'BindSSalesProcesses':
                        $fieldAlias = $this->getAliasForScanner('SSalesProcesses', 'ssalesprocesses_no');
                        if($fieldAlias) {
                            $newScannerFieldsPerUser[] = $fieldAlias;
                        }
                        $newBindAction = 'BindMail';
                        if(!in_array($newBindAction, $actions)) {
                            $actions[] = $newBindAction;
                        }
                        break;
                        // case 'BindServiceContracts': // breake
                    case 'BindVendors':
                    case 'BindPartners':
                    case 'BindOSSEmployees':
                    case 'BindLeads':
                    case 'BindCompetition':
                    case 'BindContacts':
                    case 'BindAccounts':
                        $newBindAction = 'BindMail';
                        if(!in_array($newBindAction, $actions)) {
                            $actions[] = $newBindAction;
                        }
                        break;
                    default:
                        break;
                }
            }

             $foldersAll = (new \App\Db\Query())->distinct()->select(['folder','uid'])->from('vtiger_ossmailscanner_folders_uid')->where(['user_id'=>$rcId])->all();
            $folders = array_unique(array_column($foldersAll, 'folder'));

            $isExists = (new \App\Db\Query())->from('s_yf_mail_account')->where(['login' => $login, 'mail_server_id' => $serverId])->exists();
            if (!$isExists) {
                $db->createCommand()->insert('s_yf_mail_account', [
                    'name' => $login, 'login' => $login, 'password' => $pwd, 'status' => $status, 'mail_server_id' => $serverId,
                    'scanner_actions' => implode(',', $actions), 'scanner_fields' => implode(',', $newScannerFieldsPerUser), 'folders' => \App\Json::encode($folders),
                    'owner_id' => $crm_user_id,'private' => $private
                ])->execute();

                $accountId = $db->getLastInsertID('s_yf_mail_account_id_seq');
                $foldersAll = array_intersect_key($foldersAll, $folders);
                foreach ($foldersAll as $folder) {
                    $db->createCommand()->insert('u_yf_mailscanner_folders', ['id' => $accountId, 'name' => $folder['folder'], 'uid' => $folder['uid']])->execute();
                }

                // update maile message with new mail account
                $db->createCommand()->update('vtiger_ossmailview', ['rc_user' => $accountId], ['rc_user' => $rcId])->execute();
                $db->createCommand()->update('roundcube_users_autologin', ['rcuser_id' => $accountId], ['rcuser_id' => $rcId])->execute();
            }
        }
        $dataReader->close();

        $db->createCommand("DELETE FROM `roundcube_users_autologin` WHERE `rcuser_id` NOT IN (SELECT id FROM `s_yf_mail_account`);")->execute();

        $importerBase = new \App\Db\Importers\Base();
        $importerBase->foreignKey = [
            ['roundcube_users_autologin_ibfk_1', 'roundcube_users_autologin', 'rcuser_id', 's_#__mail_account', 'id', 'CASCADE', null],
        ];

        $this->importer->updateForeignKey($importerBase);

        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s') . ' | ' . round((microtime(true) - $start) / 60, 2) . ' mim.');
    }

    private function getAliasForScanner(string $moduleName, string $fieldName)
    {
        $fieldAlias = '';
        $field = (new \App\Db\Query())->select(['fieldid', 'uitype'])->from('vtiger_field')->where(['tabid'=>\App\Module::getModuleId($moduleName),'fieldname'=>$fieldName])->one();
        if($field) {
            $fieldId = $field['fieldid'];
            $uiType = $field['uitype'];
            $fieldAlias = "{$fieldId}|{$moduleName}|{$fieldName}|{$uiType}";
        }

        return $fieldAlias;
    }
    public function smtp()
    {
        $start = microtime(true);
        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s'));
        $db = \App\Db::getInstance();
        $db->createCommand()->update('s_yf_mail_smtp', ['auth_method' => 'basic'])->execute();

        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s') . ' | ' . round((microtime(true) - $start) / 60, 2) . ' mim.');
    }

    private function addActionMapping()
    {
        $start = microtime(true);
        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s'));

        $modules = array_map('\App\Module::getModuleId', ['Calendar','OSSTimeControl','Reservations','Occurrences']);
        $this->actionMapp([
            ['type' => 'add', 'name' => 'CalendarExtraSources', 'tabsData' => $modules, 'permission' => 1],
            ['type' => 'add', 'name' => 'CalendarExtraSourcesCreate', 'tabsData' => $modules, 'permission' => 1],
            ['type' => 'remove', 'name' => 'ReloadRelationRecord'],
        ]);

        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s') . ' | ' . round((microtime(true) - $start) / 60, 2) . ' min');
    }

    private function actionMapp(array $actions)
    {
        $start = microtime(true);
        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s'));

        $db = \App\Db::getInstance();

        foreach ($actions as $action) {
            $key = (new \App\Db\Query())->select(['actionid'])->from('vtiger_actionmapping')->where(['actionname' => $action['name']])->limit(1)->scalar();
            if ('remove' === $action['type']) {
                if ($key) {
                    $db->createCommand()->delete('vtiger_actionmapping', ['actionid' => $key])->execute();
                    $db->createCommand()->delete('vtiger_profile2utility', ['activityid' => $key])->execute();
                }
                continue;
            }
            if (empty($key)) {
                $securitycheck = 0;
                $key = $db->getUniqueID('vtiger_actionmapping', 'actionid', false);
                $db->createCommand()->insert('vtiger_actionmapping', ['actionid' => $key, 'actionname' => $action['name'], 'securitycheck' => $securitycheck])->execute();
            }
            $permission = 1;
            if (isset($action['permission'])) {
                $permission = $action['permission'];
            }

            $tabsData = $action['tabsData'];
            $dataReader = (new \App\Db\Query())->select(['profileid'])->from('vtiger_profile')->createCommand()->query();
            while (false !== ($profileId = $dataReader->readColumn(0))) {
                foreach ($tabsData as $tabId) {
                    $isExists = (new \App\Db\Query())->from('vtiger_profile2utility')->where(['profileid' => $profileId, 'tabid' => $tabId, 'activityid' => $key])->exists();
                    if (!$isExists) {
                        $db->createCommand()->insert('vtiger_profile2utility', [
                            'profileid' => $profileId, 'tabid' => $tabId, 'activityid' => $key, 'permission' => $permission,
                        ])->execute();
                    }
                }
            }
            $dataReader->close();
        }
        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s') . ' | ' . round((microtime(true) - $start) / 60, 2) . ' mim.');
    }



    /**
     * Add fields.
     *
     * @param mixed $fields
     */
    public function addFields($fields = [])
    {
        $start = microtime(true);
        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s'));

        foreach ($fields as $field) {
            $moduleName = $field['moduleName'];
            $moduleId = \App\Module::getModuleId($moduleName);
            if (!$moduleId) {
                $this->log("  [ERROR] Module not exists: {$moduleName}");
                continue;
            }
            $isExists = (new \App\Db\Query())->from('vtiger_field')->where(['tablename' => $field[3], 'columnname' => $field[2], 'tabid' => $moduleId])->exists();
            if ($isExists) {
                $this->log("  [INFO] Skip adding field. Module: {$moduleName}({$moduleId}); field name: {$field[2]}, field exists: {$isExists}");
                continue;
            }

            $blockInstance = false;
            $blockId = (new \App\Db\Query())->select(['blockid'])->from('vtiger_blocks')->where(['blocklabel' => ($field['blockData']['label'] ?? $field['blockLabel']), 'tabid' => $moduleId])->scalar();
            if ($blockId) {
                $blockInstance = \vtlib\Block::getInstance($blockId, $moduleId);
            } elseif (isset($field['blockData'])) {
                $blockInstance = new \vtlib\Block();
                foreach ($field['blockData'] as $key => $value) {
                    $blockInstance->{$key} = $value;
                }
                \Vtiger_Module_Model::getInstance($moduleName)->addBlock($blockInstance);
                $blockId = $blockInstance->id;
                $blockInstance = \vtlib\Block::getInstance($blockId, $moduleId);
            }
            if (!$blockInstance
            && !($blockInstance = reset(Vtiger_Module_Model::getInstance($moduleName)->getBlocks()))) {
                $this->log("  [ERROR] No block found to create a field, you will need to create a field manually.
				Module: {$moduleName}, field name: {$field[6]}, field label: {$field[7]}");
                \App\Log::error("No block found ({$field['blockData']['label']}) to create a field, you will need to create a field manually.
				Module: {$moduleName}, field name: {$field[6]}, field label: {$field[7]}");
                continue;
            }
            $fieldInstance = new \vtlib\Field();
            $fieldInstance->column = $field[2];
            $fieldInstance->name = $field[6];
            $fieldInstance->label = $field[7];
            $fieldInstance->table = $field[3];
            $fieldInstance->uitype = $field[5];
            $fieldInstance->typeofdata = $field[15];
            $fieldInstance->readonly = $field[8];
            $fieldInstance->displaytype = $field[14];
            $fieldInstance->masseditable = $field[19];
            $fieldInstance->quickcreate = $field[16];
            $fieldInstance->columntype = $field['type'];
            $fieldInstance->presence = $field[9];
            $fieldInstance->maximumlength = $field[11];
            $fieldInstance->quicksequence = $field[17];
            $fieldInstance->info_type = $field[18];
            $fieldInstance->helpinfo = $field[20];
            $fieldInstance->summaryfield = $field[21];
            $fieldInstance->generatedtype = $field[4];
            $fieldInstance->defaultvalue = $field[10];
            $fieldInstance->fieldparams = $field[22];
            $blockInstance->addField($fieldInstance);
            if (!empty($field['picklistValues']) && (15 == $field[5] || 16 == $field[5] || 33 == $field[5])) {
                $fieldInstance->setPicklistValues($field['picklistValues']);
            }
            if (!empty($field['relatedModules']) && 10 == $field[5]) {
                $fieldInstance->setRelatedModules($field['relatedModules']);
            }
        }
        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s') . ' | ' . round((microtime(true) - $start) / 60, 2) . ' mim.');
    }

    /**
     * Post update.
     */
    public function createConfigFiles(): bool
    {
        $start = microtime(true);
        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s'));

        $files = ['app/ConfigFile.php','config/ConfigTemplates.php', 'config/Components/ConfigTemplates.php', 'modules/OSSMail/ConfigTemplate.php', 'modules/OpenStreetMap/ConfigTemplate.php', 'modules/ModComments/ConfigTemplate.php', 'modules/ReportTemplate/ConfigTemplate.php'];
        $dataReader = (new \App\Db\Query())->select(['name'])->from('vtiger_tab')->createCommand()->query();
        while ($moduleName = $dataReader->readColumn(0)) {
            $filePath = 'modules' . \DIRECTORY_SEPARATOR . $moduleName . \DIRECTORY_SEPARATOR . 'ConfigTemplate.php';
            if (file_exists($filePath)) {
                $files[] = $filePath;
            }
        }

        $filePath = 'modules/ReportTemplate';
        if(!file_exists($filePath)){
            mkdir($filePath, 0755);
        }

        $dataReader->close();
        foreach ($files as $configTemplates) {
            if(!file_exists(__DIR__ . '/files/' . $configTemplates)) {
                continue;
            }
            $path = ROOT_DIRECTORY . '/' . $configTemplates;
            copy(__DIR__ . '/files/' . $configTemplates, $path);
            \App\Cache::resetFileCache($path);
        }

        \App\Cache::resetOpcache();
        clearstatcache();

        // aouth2
        $services = \App\Config::api('enabledServices') ?: [];
        if(!in_array('webservice', $services)) {
            $services[] = 'webservice';
            $configFile = new \App\ConfigFile('api');
            $configFile->set('enabledServices', $services);
            $configFile->create();
        }
        $cookieSameSite = \App\Config::security('cookieSameSite') ?: '';
        if('Strict' === $cookieSameSite) {
            $configFile = new \App\ConfigFile('security');
            $configFile->set('cookieSameSite', 'Lax');
            $configFile->create();
        }
        $customFile =  \App\Config::performance('LOAD_CUSTOM_FILES');
        if(!$customFile) {
            $configFile = new \App\ConfigFile('performance');
            $configFile->set('LOAD_CUSTOM_FILES', true);
            $configFile->create();
        }
        $allowedImageDomains =  \App\Config::security('allowedImageDomains') ?: [];
        if(!$allowedImageDomains) {
            $allowedImageDomains = ['*.tile.openstreetmap.org'];
            $configFile = new \App\ConfigFile('security');
            $configFile->set('allowedImageDomains', $allowedImageDomains);
            $configFile->create();
        }

        $skip = [ 'db', 'debug', 'module', 'component'];
        foreach (array_diff(\App\ConfigFile::TYPES, $skip) as $type) {
            (new \App\ConfigFile($type))->create();
        }

        $dataReader = (new \App\Db\Query())->select(['name'])->from('vtiger_tab')->createCommand()->query();
        while ($moduleName = $dataReader->readColumn(0)) {
            $filePath = 'modules' . \DIRECTORY_SEPARATOR . $moduleName . \DIRECTORY_SEPARATOR . 'ConfigTemplate.php';
            if (file_exists($filePath)) {
                (new \App\ConfigFile('module', $moduleName))->create();
            }
        }
        $dataReader->close();

        $path = \ROOT_DIRECTORY . \DIRECTORY_SEPARATOR . 'config' . \DIRECTORY_SEPARATOR . 'Components' . \DIRECTORY_SEPARATOR . 'ConfigTemplates.php';
        $componentsData = require_once "$path";
        foreach ($componentsData as $component => $data) {
            (new \App\ConfigFile('component', $component))->create();
        }

        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s') . ' | ' . round((microtime(true) - $start) / 60, 2) . ' mim.');
        return true;
    }

    /**
     * Post update .
     */
    public function postupdate()
    {
        $start = microtime(true);
        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s'));

        (new \App\BatchMethod(['method' => '\App\Db\Fixer::baseModuleTools', 'params' => []]))->save();
        (new \App\BatchMethod(['method' => '\App\Db\Fixer::baseModuleActions', 'params' => []]))->save();
        (new \App\BatchMethod(['method' => '\App\Db\Fixer::profileField', 'params' => []]))->save();
        (new \App\BatchMethod(['method' => '\App\UserPrivilegesFile::recalculateAll', 'params' => []]))->save();
        (new \App\BatchMethod(['method' => 'Settings_SharingAccess_Module_Model::recalculateSharingRules', 'params' => []]))->save();

        \App\Cache::clear();
        \App\Cache::resetOpcache();
        if ($this->error || false !== strpos($this->importer->logs, 'Error')) {
            $this->stopProcess();
            $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s') . ' | ' . round((microtime(true) - $start) / 60, 2) . ' min');
            exit;
        } else {

            $db= \App\Db::getInstance();
            $db->createCommand()->insert('yetiforce_updates', [
                'user' => \App\User::getCurrentUserModel()->getDetail('user_name'),
                'name' => (string) $this->moduleNode->label,
                'from_version' => (string) $this->moduleNode->from_version,
                'to_version' => (string) $this->moduleNode->to_version,
                'result' => true,
                'time' => date('Y-m-d H:i:s'),
            ])->execute();

            $db->createCommand()->update('vtiger_version', ['current_version' => (string) $this->moduleNode->to_version])->execute();

            \vtlib\Functions::recurseDelete('cache/updates/updates');
            register_shutdown_function(function () {
                try {
                    \vtlib\Functions::recurseDelete('cache/templates_c');
                } catch (\Exception $e) {
                    \App\Log::error($e->getMessage() . PHP_EOL . $e->__toString());
                }
            });
            \App\Cache::clear();
            \App\Cache::clearOpcache();
            \vtlib\Functions::recurseDelete('app_data/LanguagesUpdater.json');
            \vtlib\Functions::recurseDelete('app_data/SystemUpdater.json');
            \vtlib\Functions::recurseDelete('app_data/cron.php');
            \vtlib\Functions::recurseDelete('app_data/ConfReport_AllErrors.php');
            \vtlib\Functions::recurseDelete('app_data/shop.php');
            \vtlib\Functions::recurseDelete('app_data/shopCache.php');
            file_put_contents('cache/logs/update.log', PHP_EOL . date('Y-m-d H:i:s') . ' | ' . ob_get_clean(), FILE_APPEND);
            $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s') . ' | ' . round((microtime(true) - $start) / 60, 2) . ' min');
            if(PHP_SAPI !== 'cli') {
                header('location: index.php?parent=Settings&module=Updates&view=Index');
            }
        }
        exit;
    }

    /**
     * Stop process when an error occurs.
     */
    public function stopProcess()
    {
        $start = microtime(true);
        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s'));
        try {
            $dbCommand = \App\Db::getInstance()->createCommand();
            $dbCommand->insert('yetiforce_updates', [
                'user' => \Users_Record_Model::getCurrentUserModel()->get('user_name'),
                'name' => (string) $this->moduleNode->label,
                'from_version' => (string) $this->moduleNode->from_version,
                'to_version' => (string) $this->moduleNode->to_version,
                'result' => false,
                'time' => date('Y-m-d H:i:s'),
            ])->execute();
            $dbCommand->update('vtiger_version', ['current_version' => (string) $this->moduleNode->to_version])->execute();
            \vtlib\Functions::recurseDelete('cache/updates');
            \vtlib\Functions::recurseDelete('cache/templates_c');

            \App\Cache::clear();
            \App\Cache::clearOpcache();
            clearstatcache();
        } catch (\Throwable $ex) {
            file_put_contents('cache/logs/update.log', $ex->__toString(), FILE_APPEND);
        }
        $logs = '';
        if ($this->error) {
            $logs = '<blockquote style="font-size: 14px;background: #EDEDED;padding: 10px;white-space: pre-line;margin-top: 10px;">' . implode(PHP_EOL, $this->error) . '</blockquote>';
        }

        file_put_contents('cache/logs/update.log', ob_get_contents(), FILE_APPEND);
        ob_end_clean();
        echo '<div class="modal in" style="display: block;overflow-y: auto;top: 30px;"><div class="modal-dialog" style="max-width: 80%;"><div class="modal-content" style="-webkit-box-shadow: inset 2px 2px 14px 1px rgba(0,0,0,0.75);-moz-box-shadow: inset 2px 2px 14px 1px rgba(0,0,0,0.75);box-shadow: inset 2px 2px 14px 1px rgba(0,0,0,0.75);-webkit-box-shadow: 2px 2px 14px 1px rgba(0,0,0,0.75);
    -moz-box-shadow: 2px 2px 14px 1px rgba(0,0,0,0.75);box-shadow: 2px 2px 14px 1px rgba(0,0,0,0.75);"><div class="modal-header">
		<h1 class="modal-title"><span class="fas fa-skull-crossbones mr-2"></span>' . \App\Language::translate('LBL__UPDATING_MODULE', 'Settings:ModuleManager') . '</h1>
		</div><div class="modal-body" style="font-size: 27px;">Some errors appeared during the update.
		We recommend verifying logs and updating the system once again.' . $logs . '<blockquote style="font-size: 14px;background: #EDEDED;padding: 10px;white-space: pre-line;">' . $this->importer->logs . '</blockquote></div><div class="modal-footer">
		<a class="btn btn-success" href="' . \App\Config::main('site_URL') . '"><span class="fas fa-home mr-2"></span>' . \App\Language::translate('LBL_HOME') . '<a>
		</div></div></div></div>';

        $this->log(__METHOD__ . ' | ' . date('Y-m-d H:i:s') . ' | ' . round((microtime(true) - $start) / 60, 2) . ' min');
        exit;
    }
}
